#include "mainMDF.h"
#include "lecfi.h"
#include "ecrfi.h"
#include "nlopt.h"


/*************************/
/*                       */
/*Penalized MDF Program  */
/*                       */
/************************/


typedef struct {
  int nbIndividus;
  int nbVariables;
  distance_t *distancesReference;
  double **target;
} functionData_t;

double myfunc (unsigned n,const double *x, double *grad, void *my_func_data)
{
   int i,j,k,ind1,ind2,a;
   double temp,distanceCarre,sommeObj,distance,compteurDistance,somme,objectifFunction,lambda,s,gamma;
   int nbIndividus,nbVariables,NbDisplacement;
   double **target;
   distance_t *distancesReference;
   functionData_t *functionData=(functionData_t *)my_func_data;
 
   
   nbIndividus=functionData->nbIndividus;
   nbVariables=functionData->nbVariables; 
   target= functionData->target;
   distancesReference=functionData->distancesReference; 
   sommeObj=0;
   somme=0;
   lambda=0.5*nbIndividus;
   gamma=0.5;
   compteurDistance=0;
   NbDisplacement=0;
  for(k=1;k<nbIndividus;k++){
   compteurDistance=compteurDistance+(nbIndividus-k);
  }
 	

  for(i=0;i<compteurDistance;i++){
     	ind1=distancesReference[i].i;
     	ind2=distancesReference[i].j;
     	distance=distancesReference[i].distance;
        distanceCarre=0;
        for(k=0;k<nbVariables;k++) {
         temp=target[ind1][k]-target[ind2][k]+x[ind1*nbVariables+k]-x[ind2*nbVariables+k];
	 distanceCarre+=temp*temp;
        }
        
        sommeObj+= ((distancesReference[i].distance)-sqrt(distanceCarre))*((distancesReference[i].distance)-sqrt(distanceCarre));   
     }
  for(i=0;i<nbIndividus*nbVariables;i++){
      s=x[i]*x[i];
     somme+=sqrt(s);
    }
   for(i=0;i<nbIndividus*nbVariables;i++){
      if(x[i]!=0){
	     NbDisplacement=NbDisplacement+1;
		 }
    }

    objectifFunction=sommeObj+lambda*gamma*somme+lambda*(1-gamma)*NbDisplacement;
    
    return objectifFunction;

}

/************************************/
/*                                  */
/*  Optimization problem by NLOPT   */
/*                                  */
/************************************/

int main(int argc,char **argv) {
  int i,j,k,l,nbIndividus,nbVariables,nbIterations;
  int n,compteurDistance;  

   
  /*Target matrix */
  double **targetResultat;
   double **target;
  /*Reference matrix*/
  distance_t *distancesReference;

  char *nomFichierSortie,*nomFichierTarget,*nomFichierReference,*option;

  srand(getpid());

  if(argc<5) {
    printf("Usage : MDF targetFile referenceFile exitFile nbIterations [--natural|--real]\n");
    exit(0);
  }   
  nomFichierTarget=strdup(argv[1]);
  nomFichierReference=strdup(argv[2]);
  nomFichierSortie=strdup(argv[3]);
  nbIterations=atoi(argv[4]);

  lectureDonnees(argv[1],argv[2],&nbIndividus,&nbVariables,&target,&distancesReference,compteurDistance);

  printf("nbPoints : %d, nbVariables : %d\n",nbIndividus,nbVariables);

  /*allocation memoire*/
  targetResultat=(double **)malloc(sizeof(double)*nbIndividus);
  for(i=0;i<nbIndividus;i++) {
    targetResultat[i]=(double *)malloc(sizeof(double)*nbVariables);
  }
  /*fin allocation memoire*/       
        
 
 functionData_t functionData;
 functionData.nbIndividus=nbIndividus;
 functionData.nbVariables=nbVariables;
 functionData.distancesReference=distancesReference;
 functionData.target=target;


   /* minimum, maximum calculation */
double xmax, ymax,zmax;
xmax=0;
ymax=0;
zmax=0;
for(i=0;i<nbIndividus;i++) {
   if (xmax<=target[i][0]) {
     xmax=target[i][0];
    }
   if (ymax<=target[i][1]) {
     ymax=target[i][1];
    } 
   if (zmax<=target[i][2]) {
     zmax=target[i][2];
    }
}
double xmin, ymin,zmin;
xmin=target[0][0];
ymin=target[0][1];
zmin=target[0][2];
for(i=0;i<nbIndividus;i++) {
   if (xmin>=target[i][0]) {
     xmin=target[i][0];
    }
   if (ymin>target[i][1]) {
     ymin=target[i][1];
    } 
   if (zmin>=target[i][2]) {
     zmin=target[i][2];
    }
}
double *min;
min= (double *)malloc(sizeof(double)*3); 
double *max;
max=(double *)malloc(sizeof(double)*3);
min[0]=xmin;
min[1]=ymin;
min[2]=zmin;
max[0]=xmax;
max[1]=ymax;
max[2]=zmax;
    
    /*Upper and lower bounds */
       n=nbIndividus*nbVariables;
       double *lb;
       lb= (double *)malloc(sizeof(double)*n); 
       double *ub;
       ub=(double *)malloc(sizeof(double)*n);
         
      for(i=0;i<nbIndividus;i++){
             lb[i*nbVariables] =-(xmax-xmin); /*lower bounds for X*/
             ub[i*nbVariables]=(xmax-xmin); /*upper bounds for X*/
             lb[i*nbVariables+1] =-(ymax-ymin); /*lower bounds for Y*/
             ub[i*nbVariables+1]=(ymax-ymin); /*upper bounds for Y*/
             lb[i*nbVariables+2] =-(zmax-zmin); /*lower bounds for Z*/
             ub[i*nbVariables+2]=(zmax-zmin); /*upper bounds for Z*/
         }
     
    /* GLOBALE OPTIMISATION */
    nlopt_opt opt;
        opt = nlopt_create(NLOPT_GN_DIRECT_L, n); /* algorithm and dimensionality */
       
    nlopt_set_upper_bounds(opt, ub);
 	nlopt_set_lower_bounds(opt, lb);

        /* Objectif function */
       double *x;
        x=(double *)malloc(sizeof(double)*n);

        nlopt_set_min_objective(opt, myfunc,&functionData);
       /* The tolerance */
      nlopt_set_xtol_rel(opt, 1.0);
       /*Initialization*/
     for(k=0;k<n;k++){
       x[k]=0;
      }
       double minfGlobal; /* the minimum objective value, upon return */
       if (nlopt_optimize(opt, x, &minfGlobal) < 0) {
             printf("nlopt failed!\n");
          }
        else {
          
           for(i=0;i<nbIndividus;i++) {
		for(j=0;j<nbVariables;j++) { 
              		targetResultat[i][j]=x[i*nbVariables+j];        
              }
            }
         printf("minfGlobal : %f\n",minfGlobal);
      }
    
     /*OPTIMISATION LOCALE*/
        nlopt_opt opt1;
        opt1= nlopt_create(NLOPT_LN_SBPLX, n); /* algorithm and dimensionality */
       
        nlopt_set_upper_bounds(opt1, ub);
 	nlopt_set_lower_bounds(opt1, lb);

        /* Objectif function */
       double *x1;
        x1=(double *)malloc(sizeof(double)*n);

        nlopt_set_min_objective(opt1, myfunc,&functionData);

       /* The tolerance */
      nlopt_set_xtol_rel(opt1, 1.0);
      

   k=0;  
       for(i=0;i<nbIndividus;i++){
           for(j=0;j<nbVariables;j++){ 
            x1[k]=targetResultat[i][j]; /* some initial guess */          
             k=k+1;   
       } 
}

        double minfLocal; /* the minimum objective value, upon return */
       if (nlopt_optimize(opt1, x1, &minfLocal) < 0) {
             printf("nlopt failed!\n");
          }
        else {
          
           for(i=0;i<nbIndividus;i++) {
		for(j=0;j<nbVariables;j++) { 
              		targetResultat[i][j]=x1[i*nbVariables+j]/*+target[i][j]*/; 
             }
            }
         printf("minfLocal : %f\n",minfLocal);
      }
    
  ecritureMatriceTargetResultat(nomFichierSortie,nbIndividus,nbVariables,targetResultat);
  
  /*desallocation memoire*/
 for(i=0;i<nbIndividus;i++) {
     free(target[i]);
   }

 free(target);
 free(lb);
 free(ub);
 free(x1);
free(x);
free(min);
free(max);
  
  /*fin desallocation memoire*/

  exit(0);


}
